# Amethyst - An experimental Minecraft server written in Elixir.
# Copyright (C) 2024 KodiCraft
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

defmodule Amethyst.DataGenerator do
  require Logger

  @moduledoc """
  This module allows Amethyst to download the vanilla Minecraft server and generate
  necessary data files.
  """
  @spec get_version_meta(String.t() | :latest) :: {:ok, map()} | {:error, term}
  def get_version_meta(version) do
    version = if version == :latest, do: get_latest_version(), else: {:ok, version}
    case version do
      {:ok, version} ->
        case Req.get(Application.fetch_env!(:amethyst, :mojang_game_meta)) do
          {:ok, %{body: %{"versions" => versions}}} ->
            case Enum.find(versions, fn %{"id" => id} -> id == version end) do
              nil -> {:error, :version_not_found}
              version_header ->
                # Get the version manifest
                case Req.get(version_header["url"]) do
                  {:ok, %{body: version_meta}} -> {:ok, version_meta}
                  {:error, err} -> {:error, err}
                end
            end
          {:error, err} -> {:error, err}
        end
      {:error, err} -> {:error, err}
    end
  end

  @spec get_server_jar(:latest | String.t(), path) :: {:ok, path} | {:error, term} when path: String.t()
  def get_server_jar(version, path) do
    case get_version_meta(version) do
      {:ok, %{"downloads" => %{"server" => %{"url" => url, "sha1" => sha1}}}} ->
        # Check if the file already exists and has the right hash
        Logger.debug("Checking if server jar exists at #{path} with hash #{sha1}")
        case File.read(path) do
          {:ok, data} ->
            if :crypto.hash(:sha, data) |> Base.encode16() |> String.downcase() == sha1 do
              Logger.debug("Using cached server jar at #{path}")
              {:ok, path}
            else
              Logger.debug("Mismatched hash, redownloading server jar")
              download_jar(url, path)
            end
          {:error, _} -> download_jar(url, path)
        end
      {:error, err} -> {:error, err}
    end
  end

  @spec generate_data_files(String.t(), out_path, nil | String.t()) :: {:error, pos_integer()} | {:ok, out_path} when out_path: String.t()
  def generate_data_files(jar_path, out_path, java_bin \\ "java") do
    Logger.info("Generating data files")
    # mkdir the output directory since we cd into it
    File.mkdir_p!(out_path)
    case System.cmd(java_bin, ["-DbundlerMainClass=net.minecraft.data.Main", "-jar", jar_path, "--all", "--output", out_path], cd: out_path) do
      {_, 0} -> {:ok, out_path}
      {_, code} -> {:error, code}
    end
  end

  @spec populate_blockregistry(String.t()) :: :ok | {:error, term}
  def populate_blockregistry(data_path) do
    block_file = Path.join([data_path, "reports", "blocks.json"])
    case File.read(block_file) do
      {:ok, data} ->
        case Jason.decode(data) do
          {:ok, blocks} ->
            Enum.each(blocks, fn {id, %{"states" => states}} ->
              Enum.each(states, fn
                %{"id" => bsi, "properties" => props} -> Amethyst.BlockRegistry.add(id, props, bsi)
                %{"id" => bsi, "default" => true} -> Amethyst.BlockRegistry.add(id, %{}, bsi)
              end)
            end)
            :ok
          {:error, err} -> {:error, err}
        end
      {:error, err} -> {:error, err}
    end
  end

  @spec generate_and_populate_data(String.t() | :latest, String.t(), String.t(), String.t() | nil) :: :ok | {:error, term}
  def generate_and_populate_data(version, jar_path, data_dir, java_bin \\ "java") do
    start_time = Time.utc_now()
    case get_server_jar(version, jar_path) do
      {:ok, jar_path} ->
        case generate_data_files(jar_path, data_dir, java_bin) do
          {:ok, data_path} ->
            populate_blockregistry(data_path)
            Logger.info("Finished generating and populating data in #{Time.diff(Time.utc_now(), start_time, :millisecond)}ms")
          {:error, code} -> {:error, code}
        end
      {:error, err} -> {:error, err}
    end
  end

  defp download_jar(url, path) do
    Logger.info("Downloading server jar from #{url} to #{path}")
    case Req.get(url, into: File.stream!(path)) do
      {:ok, _} -> {:ok, path}
      {:error, err} -> {:error, err}
    end
  end

  @spec get_latest_version() ::
          {:error, %{:__exception__ => true, :__struct__ => atom(), optional(atom()) => any()}}
          | {:ok, any()}
  def get_latest_version() do
    case Req.get(Application.fetch_env!(:amethyst, :mojang_game_meta)) do
      {:ok, %{body: %{"latest" => %{"release" => release}}}} -> {:ok, release}
      {:error, err} -> {:error, err}
    end
  end
end
